# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Context (Organization / Project) managing."""

from __future__ import annotations

import uuid
from typing import Optional

import periflow
from periflow.utils.fs import get_periflow_directory

org_context_path = get_periflow_directory() / "organization"
project_context_path = get_periflow_directory() / "project"


def get_current_group_id() -> Optional[uuid.UUID]:
    """Get organization ID of the current context."""
    if periflow.org_id:
        try:
            return uuid.UUID(periflow.org_id)
        except ValueError as exc:
            raise ValueError("Invalid organization ID format") from exc

    if not org_context_path.exists():
        return None

    with open(org_context_path, "r", encoding="utf-8") as f:
        group_id = uuid.UUID(f.read())
        return group_id


def set_current_group_id(pf_group_id: uuid.UUID):
    """Set the current organization context."""
    with open(org_context_path, "w", encoding="utf-8") as f:
        f.write(str(pf_group_id))


def get_current_project_id() -> Optional[uuid.UUID]:
    """Get project ID of the current context."""
    if periflow.project_id:
        try:
            return uuid.UUID(periflow.project_id)
        except ValueError as exc:
            raise ValueError("Invalid project ID format") from exc

    if not project_context_path.exists():
        return None

    with open(project_context_path, "r", encoding="utf-8") as f:
        project_id = uuid.UUID(f.read())
        return project_id


def set_current_project_id(pf_project_id: uuid.UUID):
    """Set the current project context."""
    with open(project_context_path, "w", encoding="utf-8") as f:
        f.write(str(pf_project_id))



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# pylint: disable=redefined-builtin

"""PeriFlow Catalog CLI."""

from __future__ import annotations

import typer

from periflow.formatter import (
    JSONFormatter,
    PanelFormatter,
    TableFormatter,
    TreeFormatter,
)
from periflow.sdk.resource.catalog import Catalog as CatalogAPI
from periflow.utils.decorator import check_api
from periflow.utils.format import datetime_to_pretty_str, secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

table_formatter = TableFormatter(
    name="Catalog",
    fields=[
        "name",
        "use_count",
        "tags",
        "created_at",
    ],
    headers=[
        "Name",
        "# Uses",
        "Tags",
        "Created At",
    ],
)
panel_formatter = PanelFormatter(
    name="Overview",
    fields=[
        "id",
        "name",
        "summary",
        "description",
        "use_count",
        "created_at",
        "tags",
    ],
    headers=[
        "ID",
        "Name",
        "Summary",
        "Description",
        "# Uses",
        "Created At",
        "Tags",
    ],
)
json_formatter = JSONFormatter(name="Attributes")
tree_formatter = TreeFormatter(name="Files")


@app.command()
@check_api
def list(
    name: str = typer.Option(
        None,
        "--name",
        "-n",
        help="The name of public checkpoint to search.",
    ),
    limit: int = typer.Option(
        20,
        "--limit",
        "-l",
        help="The number of public checkpoints to display.",
    ),
):
    """Lists public checkpoints in catalog."""
    catalogs = CatalogAPI.list(name=name, limit=limit)

    catalog_dicts = []
    for catalog in catalogs:
        catalog_dict = catalog.model_dump()
        catalog_dict["created_at"] = datetime_to_pretty_str(catalog.created_at)
        catalog_dicts.append(catalog_dict)

    table_formatter.render(catalog_dicts)


@app.command()
@check_api
def view(
    name: str = typer.Argument(),
):
    """Displays info of a catalog."""
    catalogs = CatalogAPI.list(name=name)

    catalog = None
    for cat in catalogs:
        if cat.name == name:
            catalog = cat
    if catalog is None:
        msg = f"Public checkpoint with name '{name}' is not found in the catalog."
        if len(catalogs) > 0:
            msg += f" Did you mean '{catalogs[0].name}'?"
        secho_error_and_exit(msg)

    catalog_dict = catalog.model_dump()
    catalog_dict["created_at"] = datetime_to_pretty_str(catalog.created_at)

    panel_formatter.render([catalog_dict])
    json_formatter.render(catalog_dict["attributes"])
    tree_formatter.render(catalog_dict["files"])



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Defining PeriFlow Checkpoint Converter maps."""

from __future__ import annotations

from typing import Dict, Tuple, Type, Union

from transformers import (  # type: ignore[import]
    AutoModelForCausalLM,
    BlenderbotForConditionalGeneration,
    BloomForCausalLM,
    CodeGenForCausalLM,
    FalconForCausalLM,
    GPT2LMHeadModel,
    GPTJForCausalLM,
    GPTNeoXForCausalLM,
    LlamaForCausalLM,
    OPTForCausalLM,
    PreTrainedModel,
    T5ForConditionalGeneration,
)

from periflow.errors import NotSupportedCheckpointError
from periflow.modules.converter.base import OneOfConverter
from periflow.modules.converter.models.blenderbot import BlenderbotConverter
from periflow.modules.converter.models.bloom import BloomForCausalLMConverter
from periflow.modules.converter.models.codegen import CodegenForCausalLMConverter
from periflow.modules.converter.models.falcon import FalconForCausalLMConverter
from periflow.modules.converter.models.gpt2 import GPT2LMHeadModelConverter
from periflow.modules.converter.models.gpt_neox import GPTNeoXForCausalLMConverter
from periflow.modules.converter.models.gptj import GPTJForCausalLMConverter
from periflow.modules.converter.models.llama import LlamaForCausalLMConverter
from periflow.modules.converter.models.mpt import MPTForCausalLMConverter
from periflow.modules.converter.models.opt import OPTForCausalLMConverter
from periflow.modules.converter.models.t5 import T5Converter

model_arch_converter_map: Dict[
    str, Tuple[Union[AutoModelForCausalLM, PreTrainedModel], Type[OneOfConverter]]
] = {
    "BlenderbotForConditionalGeneration": (
        BlenderbotForConditionalGeneration,
        BlenderbotConverter,
    ),
    "BloomForCausalLM": (BloomForCausalLM, BloomForCausalLMConverter),
    "CodeGenForCausalLM": (CodeGenForCausalLM, CodegenForCausalLMConverter),
    "FalconForCausalLM": (FalconForCausalLM, FalconForCausalLMConverter),
    "GPTNeoXForCausalLM": (GPTNeoXForCausalLM, GPTNeoXForCausalLMConverter),
    "GPT2LMHeadModel": (GPT2LMHeadModel, GPT2LMHeadModelConverter),
    "GPTJForCausalLM": (GPTJForCausalLM, GPTJForCausalLMConverter),
    "LlamaForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "LLaMAForCausalLM": (LlamaForCausalLM, LlamaForCausalLMConverter),
    "MPTForCausalLM": (AutoModelForCausalLM, MPTForCausalLMConverter),
    "OPTForCausalLM": (OPTForCausalLM, OPTForCausalLMConverter),
    "T5ForConditionalGeneration": (T5ForConditionalGeneration, T5Converter),
}


def get_hf_converter_factory(
    model_arch: str,
) -> Tuple[Union[AutoModelForCausalLM, PreTrainedModel], Type[OneOfConverter]]:
    """Return the converter factory for the given model architecture.

    Args:
        model_arch (str): Model architecture name.

    Returns:
        Tuple[Union[AutoModelForCausalLM, PreTrainedModel], Type[OneOfConverter]]: Tuple of
            model class and converter class.

    Raises:
        NotSupportedCheckpointError: Raised when the given model architecture is not supported.

    """
    if model_arch not in model_arch_converter_map:
        raise NotSupportedCheckpointError(
            invalid_option=f"Model architecture='{model_arch}'",
            valid_options=list(model_arch_converter_map.keys()),
        )

    return model_arch_converter_map[model_arch]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Checkpoint Schemas."""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict

from periflow.enums import (
    CatalogStatus,
    CheckpointCategory,
    CheckpointStatus,
    CheckpointValidationStatus,
)


class V1Catalog(BaseModel):
    """V1 catalog schema."""

    id: UUID
    organization_id: UUID
    name: str
    attributes: Optional[Dict[str, Any]]
    tags: Optional[List[str]]
    format: str
    summary: Optional[str]
    description: Optional[str]
    use_count: int
    status: CatalogStatus
    status_reason: Optional[str]
    deleted: bool
    deleted_at: Optional[datetime]
    files: List[V1CheckpointFile]
    created_at: datetime


class V1CheckpointFile(BaseModel):
    """V1 checkpoint file schema."""

    name: str
    path: str
    mtime: datetime
    size: int
    created_at: datetime


class V1ModelForm(BaseModel):
    """V1 model form schema."""

    id: UUID
    form_category: str
    secret_id: Optional[UUID]
    secret_type: str
    vendor: str
    region: str
    storage_name: str
    dist_json: Optional[Dict[str, Any]]
    files: List[V1CheckpointFile]
    created_at: datetime
    deleted: bool
    deleted_at: Optional[datetime]
    hard_deleted: bool
    hard_deleted_at: Optional[datetime]


class V1CheckpointOwnership(BaseModel):
    """V1 checkpoint ownership schema."""

    organization_id: UUID
    project_id: UUID


class V1Checkpoint(BaseModel):
    """V1 checkpoint schema."""

    model_config = ConfigDict(protected_namespaces=())

    id: UUID
    user_id: UUID
    model_category: CheckpointCategory
    job_id: Optional[UUID]
    name: str
    attributes: Optional[Dict[str, Any]]
    iteration: Optional[int]
    tags: Optional[List[str]]
    catalog: Optional[V1Catalog]
    status: CheckpointStatus
    status_reason: Optional[str]
    validation_status: Optional[CheckpointValidationStatus]
    validation_status_reason: Optional[str]
    created_at: datetime
    deleted: bool
    deleted_at: Optional[datetime]
    hard_deleted: bool
    hard_deleted_at: Optional[datetime]
    forms: Optional[List[V1ModelForm]] = None
    ownerships: Optional[List[V1CheckpointOwnership]] = None



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow MPT Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, Union

import numpy as np
import torch
from transformers import PretrainedConfig  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import DECODER_PREFIX, DecoderOnlyConverter
from periflow.modules.converter.utils import nontype_partial


def safe_config_get(config: Union[Dict[str, Any], PretrainedConfig], key: str) -> Any:
    """Safe getter from config.

    This function is a temporary function because MPT is not merged into Hugging Face's upstream yet
    (i.e., we cannot do `from transformers import MPTConfig`)
    """
    if isinstance(config, PretrainedConfig):
        config = config.to_dict()  # type: ignore

    if key not in config:
        raise CheckpointConversionError(f"{key} does not exist in the config")

    return config[key]


class MPTForCausalLMConverter(DecoderOnlyConverter):
    """MPTForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if MPT architectures' config can be converted to PeriFlow format."""
        super().check_config()
        attn_config = safe_config_get(self.config, "attn_config")

        if not safe_config_get(attn_config, "alibi"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'alibi={safe_config_get(attn_config, 'alibi')}'",
                valid_options=[True],
            )

        if safe_config_get(attn_config, "alibi_bias_max") != 8:
            raise NotSupportedCheckpointError(
                invalid_option=f"'alibi={safe_config_get(attn_config, 'alibi_bias_max')}'",
                valid_options=[8],
            )

        if safe_config_get(attn_config, "attn_type") != "multihead_attention":
            raise NotSupportedCheckpointError(
                invalid_option=f"'attn_type={safe_config_get(attn_config, 'attn_type')}'",
                valid_options=["multihead_attention"],
            )

        if safe_config_get(attn_config, "prefix_lm"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'prefix_lm={safe_config_get(attn_config, 'prefix_lm')}'",
                valid_options=[False],
            )

        if safe_config_get(attn_config, "qk_ln"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'qk_ln={safe_config_get(attn_config, 'qk_ln')}'",
                valid_options=[False],
            )

        if safe_config_get(attn_config, "softmax_scale") is not None:
            raise NotSupportedCheckpointError(
                invalid_option=f"'softmax_scale={safe_config_get(attn_config, 'softmax_scale')}'",
                valid_options=[None],
            )

        if safe_config_get(self.config, "expansion_ratio") != 4:
            raise NotSupportedCheckpointError(
                invalid_option=(
                    f"'expansion_ratio={safe_config_get(self.config, 'expansion_ratio')}'"
                ),
                valid_options=[4],
            )

        if not safe_config_get(self.config, "no_bias"):
            raise NotSupportedCheckpointError(
                invalid_option=f"'no_bias={safe_config_get(self.config, 'no_bias')}'",
                valid_options=[True],
            )

        if safe_config_get(self.config, "logit_scale") is not None:
            raise NotSupportedCheckpointError(
                invalid_option=(
                    f"'logit_scale={safe_config_get(self.config, 'logit_scale')}'"
                ),
                valid_options=[None],
            )

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in MPT."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".norm_1.weight"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".norm_2.weight"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[".attn.Wqkv.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".attn.out_proj.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".ffn.up_proj.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".ffn.down_proj.weight"],
            ),
        }

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in MPT."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["transformer.wte.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["transformer.norm_f.weight"],
            ),
        }

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The MPT model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            safe_config_get(self.config, "max_seq_len"),
        )

        attn_config = safe_config_get(self.config, "attn_config")

        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": safe_config_get(self.config, "max_seq_len"),
            "vocab_size": safe_config_get(self.config, "vocab_size"),
            "clip_qkv": safe_config_get(attn_config, "clip_qkv") or 0.0,
            "eos_token": self.get_eos_token_id() or "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "mpt"

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before the MPT's transformer block number."""
        return "transformer.blocks."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in Falcon."""
        return safe_config_get(self.config, "n_layers")

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in Falcon."""
        return safe_config_get(self.config, "d_model")

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in Falcon."""
        return safe_config_get(self.config, "n_heads")

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in Falcon."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of Falcon."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow LLaMAa Checkpoint Converter."""


from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import LlamaConfig  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import DECODER_PREFIX, DecoderOnlyConverter
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    convert_to_gpt_j_params,
    get_tensor_from_state_dict,
    nontype_partial,
)


class LlamaForCausalLMConverter(DecoderOnlyConverter):
    """LlamaForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if LLaMA architectures' config can be converted to PeriFlow format."""
        super().check_config()
        try:
            if cast(LlamaConfig, self.config).hidden_act not in ["silu"]:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'hidden_act={cast(LlamaConfig, self.config).hidden_act}'",
                    valid_options=["silu"],
                )
            if cast(LlamaConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(LlamaConfig, self.config).rms_norm_eps not in (1e-5, 1e-6):
                raise NotSupportedCheckpointError(
                    invalid_option=f"'rms_norm_eps={cast(LlamaConfig, self.config).rms_norm_eps}'",
                    valid_options=[1e-5, 1e-6],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],  # type: ignore[override]
    ) -> np.ndarray:
        """qkv_weight_convert for LLaMA's attention layer."""
        assert len(per_layer_postfixes) == 3
        q_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        k_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[1]
        )
        v_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[2]
        )

        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)
        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        qkv_weight = torch.cat([q_weight, k_weight, v_weight], dim=0)
        qkv_weight = qkv_weight.transpose(0, -1)
        return convert_tensor_to_np_array(qkv_weight, self.data_type)

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(LlamaConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The Llama model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "ff_intermediate_size": config.intermediate_size,
            "max_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "llama"

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in LLaMA."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".input_layernorm.weight"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[
                    ".self_attn.q_proj.weight",
                    ".self_attn.k_proj.weight",
                    ".self_attn.v_proj.weight",
                ],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".self_attn.o_proj.weight"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".post_attention_layernorm.weight"],
            ),
            "mlp/c_gate/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.gate_proj.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.up_proj.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.down_proj.weight"],
            ),
        }

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in LLaMA."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["model.embed_tokens.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["model.norm.weight"],
            ),
            "head_fc/weight:0": nontype_partial(
                self.head_weight_convert,
                per_layer_postfixes=["lm_head.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before LLaMA's transformer block number."""
        return "model.layers."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in LLaMA."""
        return cast(LlamaConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in LLaMA."""
        return cast(LlamaConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in LLaMA."""
        return cast(LlamaConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in LLaMA."""
        config = cast(LlamaConfig, self.config)
        if config.num_key_value_heads is None:
            return self.decoder_num_attention_heads
        return config.num_key_value_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of LLaMA."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimension of LLaMA."""
        return self.decoder_head_size



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow T5 Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import T5Config  # type: ignore[import]

from periflow.enums import CheckpointDataType
from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import (
    DECODER_PREFIX,
    ENCODER_PREFIX,
    EncoderDecoderConverter,
)
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    get_tensor_from_state_dict,
    nontype_partial,
)


class T5Converter(EncoderDecoderConverter):
    """T5ForConditionalGeneration Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if T5 architectures' config can be converted to PeriFlow format."""
        super().check_config()
        try:
            if not (
                cast(T5Config, self.config).is_gated_act
                ^ cast(T5Config, self.config).tie_word_embeddings
            ):
                raise NotSupportedCheckpointError(
                    invalid_option=f"'is_gated_act={cast(T5Config, self.config).is_gated_act}'and "
                    f"'tie_word_embeddings={cast(T5Config, self.config).tie_word_embeddings}'",
                    valid_options=[
                        "'is_gated_act' and 'tie_word_embeddings' should be different."
                    ],
                )

            if cast(T5Config, self.config).layer_norm_epsilon != 1e-6:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_epsilon="
                    f"{cast(T5Config, self.config).layer_norm_epsilon}'",
                    valid_options=[1e-6],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def _decoder_final_ln_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Special handle for T5."""
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])

        if cast(T5Config, self.config).tie_word_embeddings:
            param = param * (cast(T5Config, self.config).d_model ** -0.5)

        return convert_tensor_to_np_array(param, self.data_type)

    def pos_embed_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """Convert positional embedding weights in T5."""
        assert len(per_layer_postfixes) == 1
        param = get_tensor_from_state_dict(state_dict, layer + per_layer_postfixes[0])
        return convert_tensor_to_np_array(
            param=param, data_type=CheckpointDataType.FP32
        )

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(T5Config, self.config)

        logger.warn(
            "The 'max_input_length' and 'max_output_length' fields are left blank as "
            "they cannot be automatically configured. "
            "Determine the 'max_input_length' and 'max_output_length' according to your "
            "needs. The T5 model does not rely on absolute position embeddings, "
            "allowing you to choose any suitable value."
        )

        eos_token_id = self.get_eos_token_id()
        decoder_start_token_id = self.get_decoder_start_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.encoder_head_size,
            "num_heads": self.encoder_num_attention_heads,
            "hidden_size": self.encoder_hidden_size,
            "ff_intermediate_size": config.d_ff,
            "num_encoder_layers": self.encoder_layer_num,
            "num_decoder_layers": self.decoder_layer_num,
            "max_input_length": "FILL ME",
            "max_output_length": "FILL ME",
            "num_pos_emb_buckets": config.relative_attention_num_buckets,
            "max_pos_distance": config.relative_attention_max_distance,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
            "decoder_start_token": (
                decoder_start_token_id
                if decoder_start_token_id is not None
                else "FILL ME"
            ),
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        if cast(T5Config, self.config).is_gated_act:
            return "t5-v1_1"
        return "t5"

    @property
    def encoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in T5's encoder."""
        convert_dict = {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".layer.0.layer_norm.weight"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".layer.1.layer_norm.weight"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[
                    ".layer.0.SelfAttention.q.weight",
                    ".layer.0.SelfAttention.k.weight",
                    ".layer.0.SelfAttention.v.weight",
                ],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".layer.0.SelfAttention.o.weight"],
            ),
        }
        if cast(T5Config, self.config).is_gated_act:
            convert_dict.update(
                {
                    "mlp/c_gate/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.1.DenseReluDense.wi_0.weight"],
                    ),
                    "mlp/c_fc/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.1.DenseReluDense.wi_1.weight"],
                    ),
                    "mlp/c_proj/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.1.DenseReluDense.wo.weight"],
                    ),
                }
            )
        else:
            convert_dict.update(
                {
                    "mlp/c_fc/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.1.DenseReluDense.wi.weight"],
                    ),
                    "mlp/c_proj/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.1.DenseReluDense.wo.weight"],
                    ),
                }
            )
        return convert_dict

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in T5's decoder."""
        convert_dict = {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".layer.0.layer_norm.weight"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".layer.1.layer_norm.weight"],
            ),
            "ln_3/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".layer.2.layer_norm.weight"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[
                    ".layer.0.SelfAttention.q.weight",
                    ".layer.0.SelfAttention.k.weight",
                    ".layer.0.SelfAttention.v.weight",
                ],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".layer.0.SelfAttention.o.weight"],
            ),
            "cross_attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[
                    ".layer.1.EncDecAttention.q.weight",
                    ".layer.1.EncDecAttention.k.weight",
                    ".layer.1.EncDecAttention.v.weight",
                ],
            ),
            "cross_attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".layer.1.EncDecAttention.o.weight"],
            ),
        }
        if cast(T5Config, self.config).is_gated_act:
            convert_dict.update(
                {
                    "mlp/c_gate/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.2.DenseReluDense.wi_0.weight"],
                    ),
                    "mlp/c_fc/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.2.DenseReluDense.wi_1.weight"],
                    ),
                    "mlp/c_proj/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.2.DenseReluDense.wo.weight"],
                    ),
                }
            )
        else:
            convert_dict.update(
                {
                    "mlp/c_fc/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.2.DenseReluDense.wi.weight"],
                    ),
                    "mlp/c_proj/weight:0": nontype_partial(
                        self.linear_weight_convert,
                        per_layer_postfixes=[".layer.2.DenseReluDense.wo.weight"],
                    ),
                }
            )
        return convert_dict

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in T5."""
        convert_dict = {}
        convert_dict["wte/weight:0"] = nontype_partial(
            self.token_embed_weight_convert,
            per_layer_postfixes=["shared.weight"],
        )
        if not cast(T5Config, self.config).tie_word_embeddings:
            convert_dict["head_fc/weight:0"] = nontype_partial(
                self.head_weight_convert,
                per_layer_postfixes=["lm_head.weight"],
            )
        convert_dict[ENCODER_PREFIX + "/wpe/weight:0"] = nontype_partial(
            self.pos_embed_weight_convert,
            per_layer_postfixes=[
                "encoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"
            ],
        )
        convert_dict[DECODER_PREFIX + "/wpe/weight:0"] = nontype_partial(
            self.pos_embed_weight_convert,
            per_layer_postfixes=[
                "decoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"
            ],
        )
        convert_dict[ENCODER_PREFIX + "/ln_f/gamma:0"] = nontype_partial(
            self.ln_weight_convert,
            per_layer_postfixes=["encoder.final_layer_norm.weight"],
        )
        convert_dict[DECODER_PREFIX + "/ln_f/gamma:0"] = nontype_partial(
            self._decoder_final_ln_weight_convert,
            per_layer_postfixes=["decoder.final_layer_norm.weight"],
        )
        return convert_dict

    @property
    def encoder_layer_prefix(self) -> str:
        """The layer name prefix used before T5 encoder's transformer block number."""
        return "encoder.block."

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before T5 decoder's transformer block number."""
        return "decoder.block."

    @property
    def encoder_layer_num(self) -> int:
        """The number of transformer blocks in T5 encoder."""
        return cast(T5Config, self.config).num_layers

    @property
    def encoder_hidden_size(self) -> int:
        """The hidden size of T5 encoder."""
        return cast(T5Config, self.config).d_model

    @property
    def encoder_num_attention_heads(self) -> int:
        """The number of attention heads of T5 encoder."""
        return cast(T5Config, self.config).num_heads

    @property
    def encoder_head_size(self) -> int:
        """The head size of T5 encoder."""
        return cast(T5Config, self.config).d_kv

    @property
    def decoder_layer_num(self) -> int:
        """The number of transformer blocks in T5 decoder."""
        return cast(T5Config, self.config).num_decoder_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size of T5 decoder."""
        return cast(T5Config, self.config).d_model

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads of T5 decoder."""
        return cast(T5Config, self.config).num_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads of t5 decoder."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of T5 decoder."""
        return cast(T5Config, self.config).d_kv



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Checkpoint Converter Interface."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict

import h5py  # type: ignore[import]
import numpy as np
import torch
from tqdm.std import tqdm as std_tqdm


class ModelConversionInterface(ABC):
    """Interface get information for converting models."""

    @abstractmethod
    def get_convert_dict(
        self,
    ) -> Dict[str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]]:
        """Return the convert_dict for the model.

        ### convert_dict format
        convert_dict = {
            "block-type": {
                "converted_layer_name": convert_fn,
            },
        }
        """

    @abstractmethod
    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""

    @abstractmethod
    def convert(
        self,
        model: torch.nn.Module,
        output_path: str,
        state_dict: Dict[str, torch.Tensor],
        convert_dict: Dict[
            str, Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]
        ],
    ) -> None:
        """Convert Huggingface Model to PeriFlow format(.h5).

        Args:
            model (torch.nn.Module): Huggingface model.
            output_path (str): Path to save the converted checkpoint.
            state_dict (Dict[str, torch.Tensor]):
                Dictionary of mapping of tensor name to tensor
            convert_dict (Dict[Callable[[Dict[str, torch.Tensor], str], np.ndarray]]):
                Dictionary of mapping converted params name to conversion functions.

        """

    @abstractmethod
    def check_config(self) -> None:
        """Check if the model is convertable."""


class NonTFBlockConversionInterface(ABC):
    """Interface get information for converting common layers."""

    @property
    @abstractmethod
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for layers that do not belong to transformer blocks."""

    def convert_non_transformer_layers(
        self,
        state_dict: Dict[str, torch.Tensor],
        convert_dict: Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]],
        out_f: h5py.Group,
        pbar: std_tqdm,
    ) -> None:
        """Return the number of transformer blocks converted in the decoder."""
        for (
            converted_layer_name,
            convert_fn,
        ) in convert_dict.items():
            converted_np_array = convert_fn(state_dict, "")
            out_f[converted_layer_name] = converted_np_array
            pbar.update()


class DecoderTFBlockConversionInterface(ABC):
    """Interface get information for converting decoder layers."""

    @property
    @abstractmethod
    def decoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the decoder's transformer block number."""

    @property
    @abstractmethod
    def decoder_layer_num(self) -> int:
        """Return the number of transformer blocks in the decoder."""

    @property
    @abstractmethod
    def decoder_hidden_size(self) -> int:
        """Return the hidden size of the decoder."""

    @property
    @abstractmethod
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads."""

    @property
    @abstractmethod
    def decoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the decoder."""

    @property
    @abstractmethod
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for transformer blocks in the decoder."""

    @property
    @abstractmethod
    def decoder_head_size(self) -> int:
        """Return the head size of the decoder."""

    def convert_decoder_layers(
        self,
        state_dict: Dict[str, torch.Tensor],
        convert_dict: Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]],
        out_f: h5py.Group,
        pbar: std_tqdm,
    ) -> None:
        """Return the number of transformer blocks converted in the decoder."""
        for i in range(self.decoder_layer_num):
            layer = self.decoder_layer_prefix + f"{i}"
            per_layer_out_ckpt = out_f.create_group(f"h_._{i}")

            for converted_layer_name, convert_fn in convert_dict.items():
                converted_np_array = convert_fn(state_dict, layer)
                per_layer_out_ckpt[converted_layer_name] = converted_np_array
                pbar.update()


class EncoderTFBlockConversionInterface(ABC):
    """Interface get information for converting encoder layers."""

    @property
    @abstractmethod
    def encoder_layer_prefix(self) -> str:
        """Return the layer name prefix used before the encoder's transformer block number."""

    @property
    @abstractmethod
    def encoder_layer_num(self) -> int:
        """Return the number of transformer blocks in the encoder."""

    @property
    @abstractmethod
    def encoder_hidden_size(self) -> int:
        """Return the hidden size of the encoder."""

    @property
    @abstractmethod
    def encoder_num_attention_heads(self) -> int:
        """Return the number of attention heads in the encoder."""

    @property
    @abstractmethod
    def encoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """Return the convert_dict for transformer blocks in the encoder."""

    @property
    @abstractmethod
    def encoder_head_size(self) -> int:
        """Return the head size of the encoder."""

    def convert_encoder_layers(
        self,
        state_dict: Dict[str, torch.Tensor],
        convert_dict: Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]],
        out_f: h5py.Group,
        pbar: std_tqdm,
    ) -> None:
        """Return the number of transformer blocks converted in the encoder."""
        for i in range(self.encoder_layer_num):
            layer = self.encoder_layer_prefix + f"{i}"
            per_layer_out_ckpt = out_f.create_group(f"h_._{i}")

            for converted_layer_name, convert_fn in convert_dict.items():
                converted_np_array = convert_fn(state_dict, layer)
                per_layer_out_ckpt[converted_layer_name] = converted_np_array
                pbar.update()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow V1 Detokenize API Schemas."""

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class V1DetokenizeRequest(BaseModel):
    """V1 detokenize request schema."""

    tokens: Optional[List[int]] = None


class V1DetokenizeResponse(BaseModel):
    """V1 detokenize response schema."""

    text: str



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Organization CLI."""

from __future__ import annotations

from typing import Any, Dict
from uuid import UUID

import typer

from periflow.client.group import GroupClient
from periflow.client.user import UserClient, UserGroupClient
from periflow.context import get_current_group_id
from periflow.enums import GroupRole
from periflow.formatter import PanelFormatter, TableFormatter
from periflow.utils.decorator import check_api
from periflow.utils.format import secho_error_and_exit

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

org_table_formatter = TableFormatter(
    name="Organization", fields=["name", "id"], headers=["Name", "ID"]
)
org_panel_formatter = PanelFormatter(
    name="Organization Detail",
    fields=["id", "name", "status"],
    headers=["ID", "Name", "Status"],
)
member_table_formatter = TableFormatter(
    name="Members",
    fields=["id", "name", "email", "privilege_level"],
    headers=["ID", "Name", "Email", "Role"],
)


@app.command()
@check_api
def invite(email: str = typer.Argument(..., help="Invitation recipient email address")):
    """Invite a new member to the organization.

    :::info
    Only the organization **Owner** can invite members.
    The invitation email will be sent to the provided email address.
    :::
    """
    group_client = GroupClient()

    org = get_current_org()

    if org["privilege_level"] != "owner":
        secho_error_and_exit("Only the owner of the organization can invite/set-role.")

    group_client.invite_to_group(org["id"], email)
    typer.echo("Invitation Successfully Sent!")


@app.command("set-role")
@check_api
def set_role(
    email: str = typer.Argument(..., help="Email of the user to assign a role"),
    role: GroupRole = typer.Argument(..., help="Organization role"),
):
    """Set organization role of the user.

    The organization-level roles and privileges are as follows:

    |                          | Owner | Member |
    |--------------------------|-------|--------|
    | Collaborate with teams   | ✓     | ✓      |
    | Invite members           | ✓     | ✗      |
    | Assign roles             | ✓     | ✗      |
    | Create & delete projects | ✓     | ✗      |
    | Manage payments          | ✓     | ✗      |
    | Delete organization      | ✓     | ✗      |

    """
    user_client = UserClient()

    org = get_current_org()

    if org["privilege_level"] != "owner":
        secho_error_and_exit(
            "Only the owner of the organization can invite/set-privilege."
        )

    user_id = get_org_user_id_by_email(org["id"], email)
    user_client.set_group_privilege(org["id"], user_id, role)
    typer.echo(
        f"Organization role for user '{email}' successfully updated to {role.value}!"
    )


@app.command("delete-user")
@check_api
def delete_user(
    email: str = typer.Argument(
        ...,
        help="Email of the user to delete from the organization",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Forcefully delete without confirmation prompt",
    ),
):
    """Remove a user from the organization.

    :::info
    Only the **Owner** of the organization can remove a member.
    :::
    """
    user_client = UserClient()

    org = get_current_org()

    if org["privilege_level"] != "owner":
        secho_error_and_exit(
            "Only the owner of the organization can invite/set-privilege."
        )

    org_id = org["id"]
    user_id = get_org_user_id_by_email(org_id, email)

    if not force:
        do_delete = typer.confirm(
            f"Are you sure to remove user '{email}' from the organization?"
        )
        if not do_delete:
            raise typer.Abort()
    user_client.delete_from_org(user_id, org_id)

    typer.secho("User is successfully deleted from organization", fg=typer.colors.BLUE)


def get_org_user_id_by_email(org_id: UUID, email: str) -> UUID:
    """Get ID of user by the email."""
    group_client = GroupClient()
    users = group_client.list_users(org_id)
    for user in users:
        if user["email"] == email:
            return UUID(user["id"])
    secho_error_and_exit(f"User '{email}' is not a member of this organization.")


def get_current_org() -> Dict[str, Any]:
    """Get the current organization info."""
    user_group_client = UserGroupClient()

    curr_org_id = get_current_group_id()
    if curr_org_id is None:
        secho_error_and_exit("Organization is not identified. Please login again.")

    org = user_group_client.get_group_info()
    if org["id"] == str(curr_org_id):
        return org

    # org context may be wrong
    secho_error_and_exit("Failed to identify organization.")


@app.command()
@check_api
def members():
    """List up members in the current working organization."""
    group_client = GroupClient()
    org_id = get_current_group_id()

    if org_id is None:
        secho_error_and_exit("Organization is not identified. Please login again.")

    members = group_client.list_users(org_id)
    member_table_formatter.render(members)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Client: CLI and SDK for the fastest generative AI serving."""

from __future__ import annotations

import os

from periflow.di.injector import set_default_modules
from periflow.di.modules import default_modules
from periflow.schema.api.v1.completion import V1CompletionOptions
from periflow.sdk.api.completion import Completion
from periflow.sdk.api.detokenization import Detokenization
from periflow.sdk.api.tokenization import Tokenization
from periflow.sdk.init import init
from periflow.sdk.resource.checkpoint import Checkpoint
from periflow.sdk.resource.credential import Credential
from periflow.sdk.resource.deployment import Deployment

api_key = os.environ.get("PERIFLOW_API_KEY")
org_id = os.environ.get("PERIFLOW_ORG_ID")
project_id = os.environ.get("PERIFLOW_PRJ_ID")

set_default_modules(default_modules)

__all__ = [
    "api_key",
    "org_id",
    "project_id",
    "init",
    "Checkpoint",
    "Credential",
    "Deployment",
    "Completion",
    "Detokenization",
    "Tokenization",
    "V1CompletionOptions",
]



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""CLI for API Key."""

from __future__ import annotations

import typer
from dateutil.parser import parse

from periflow.client.user import UserAccessKeyClient
from periflow.formatter import PanelFormatter, TableFormatter
from periflow.utils.decorator import check_api
from periflow.utils.format import datetime_to_simple_string

app = typer.Typer(
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=False,
)

access_key_table = TableFormatter(
    name="Access Keys",
    fields=[
        "id",
        "name",
        "created_at",
    ],
    headers=[
        "ID",
        "Name",
        "Created At",
    ],
)

access_key_panel = PanelFormatter(
    name="Access Key",
    fields=[
        "id",
        "name",
        "created_at",
        "token",
    ],
    headers=[
        "ID",
        "Name",
        "Created At",
        "Access Key",
    ],
)


# pylint: disable=redefined-builtin
@app.command()
@check_api
def list():
    """List all API keys."""
    client = UserAccessKeyClient()
    access_keys = client.list_access_keys()
    for key in access_keys:
        key["created_at"] = datetime_to_simple_string(parse(key["created_at"]))
        if key["expiry_time"]:
            key["expiry_time"] = datetime_to_simple_string(parse(key["expiry_time"]))

    access_key_table.render(access_keys)


@app.command()
@check_api
def create(name: str = typer.Option(..., "--name", "-n", help="Name of api key.")):
    """Create a new API key."""
    client = UserAccessKeyClient()
    access_key = client.create_access_key(name=name)
    access_key["created_at"] = datetime_to_simple_string(
        parse(access_key["created_at"])
    )
    if access_key["expiry_time"]:
        access_key["expiry_time"] = datetime_to_simple_string(
            parse(access_key["expiry_time"])
        )

    access_key_panel.render(access_key)

    typer.secho(
        "If you misplace or cannot recall your confidential access key, "
        "there is no way to recover it.\n"
        "Your only option is to generate a fresh access key and delete the old one.",
        fg=typer.colors.BLUE,
    )


@app.command()
@check_api
def delete(access_key_id: str = typer.Argument(..., help="ID of key to delete")):
    """Delete API key."""
    client = UserAccessKeyClient()
    client.delete_access_key(access_key_id)
    typer.secho(f"Success to delete access key {access_key_id}", fg=typer.colors.BLUE)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Deployment Clients."""

from __future__ import annotations

from datetime import datetime
from string import Template
from typing import Any, Dict, List, Optional

from periflow.client.base import Client, ProjectRequestMixin


class DeploymentClient(Client[str]):
    """Deployment client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_serving_uri("deployment/"))

    def get_deployment(self, deployment_id: str) -> Dict[str, Any]:
        """Get a deployment info."""
        data = self.retrieve(
            pk=deployment_id,
        )
        return data

    def create_deployment(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """Create a new deployment."""
        data = self.post(
            json=config,
        )
        return data

    def update_deployment_scaler(
        self, deployment_id: str, min_replicas: int, max_replicas: int
    ) -> None:
        """Update a deployment auto-scaler config."""
        json_body = {
            "scaler_config": {
                "min_replica_count": min_replicas,
                "max_replica_count": max_replicas,
            },
            "update_msg": f"Set min_replicas to {min_replicas}, max_replicas to {max_replicas}",
        }
        self.partial_update(
            pk=deployment_id,
            path="scaler",
            json=json_body,
        )

    def list_deployments(
        self, project_id: Optional[str], archived: bool, limit: int, from_oldest: bool
    ) -> List[Dict[str, Any]]:
        """List all deployments."""
        params: Dict[str, Any] = {"archived": archived}
        if project_id:
            params["project_id"] = project_id
        if from_oldest:
            params["descending"] = False

        deployments = self.list(
            pagination=True,
            limit=limit,
            params=params,
        )
        return deployments

    def stop_deployment(self, deployment_id: str) -> None:
        """Delete a deployment."""
        self.delete(
            pk=deployment_id,
        )


class DeploymentLogClient(Client[str]):
    """Deployment log client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_serving_uri("deployment/$deployment_id/log/")
        )

    def get_deployment_logs(self, replica_index: int) -> List[Dict[str, Any]]:
        """Get logs from a deployment."""
        data = self.list(
            pagination=False,
            params={"replica_index": replica_index},
        )
        return data


class DeploymentMetricsClient(Client):
    """Deployment metrics client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_serving_uri("deployment/$deployment_id/metrics/")
        )

    def get_metrics(
        self, start: datetime, end: datetime, time_window: int
    ) -> List[Dict[str, Any]]:
        """Get metrics from a deployment."""
        data = self.list(
            pagination=False,
            json={
                "start": start.isoformat(),
                "end": end.isoformat(),
                "time_window_sec": time_window,
                "metrics": ["error_rates", "p50_latency_ms", "p99_latency_ms"],
            },
        )
        return data


class DeploymentEventClient(Client):
    """Deployment event client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_serving_uri("deployment/$deployment_id/event/")
        )

    def get_events(self) -> List[Dict[str, Any]]:
        """Get deployment events."""
        data = self.list(
            pagination=False,
        )
        return data


class DeploymentReqRespClient(Client):
    """Deployment request-response client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_serving_uri(
                "deployment/$deployment_id/req_resp/download/"
            )
        )

    def get_download_urls(self, start: datetime, end: datetime) -> List[Dict[str, str]]:
        """Get presigned URLs to download request-response logs."""
        params = {
            "start": start.isoformat(),
            "end": end.isoformat(),
        }
        data = self.list(
            pagination=False,
            params=params,
        )
        return data


class PFSProjectUsageClient(Client[str], ProjectRequestMixin):
    """Project-level deployment usage client for serving."""

    def __init__(self, **kwargs):
        """Initialize project usage client for serving."""
        self.initialize_project()
        super().__init__(project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_serving_uri("usage/project/$project_id/duration")
        )

    def get_project_deployment_durations(
        self,
        start_date: datetime,
        end_date: datetime,
    ) -> Dict[str, Any]:
        """Get total deployment uptime info in the project."""
        params = {
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }
        data = self.list(
            pagination=False,
            params=params,
        )
        return data


class PFSVMClient(Client):
    """VM client for serving."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_serving_uri("vm/"))

    def list_vms(self) -> List[Dict[str, Any]]:
        """List all VM info."""
        data = self.list(
            pagination=False,
        )
        return data



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Optional as _Optional

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers

DESCRIPTOR: _descriptor.FileDescriptor

class V1DetokenizeRequest(_message.Message):
    __slots__ = ["tokens"]
    TOKENS_FIELD_NUMBER: _ClassVar[int]
    tokens: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, tokens: _Optional[_Iterable[int]] = ...) -> None: ...



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Project Clients."""

# pylint: disable=too-many-arguments

from __future__ import annotations

from string import Template
from typing import Any, Dict, List, Optional
from uuid import UUID

from requests import HTTPError

from periflow.client.base import Client, ProjectRequestMixin
from periflow.enums import CredType
from periflow.errors import NotFoundError
from periflow.utils.maps import cred_type_map


def find_project_id(projects: List[Dict[str, Any]], project_name: str) -> UUID:
    """Find an ID of project among a list of projects."""
    for project in projects:
        if project["name"] == project_name:
            return UUID(project["id"])
    raise NotFoundError(f"No project exists with name {project_name}.")


class ProjectClient(Client[UUID]):
    """Project client."""

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(self.url_provider.get_auth_uri("pf_project"))

    def get_project(self, pf_project_id: UUID) -> Dict[str, Any]:
        """Get project info."""
        data = self.retrieve(pk=pf_project_id)
        return data

    def check_project_membership(self, pf_project_id: UUID) -> bool:
        """Check accessibility to the project."""
        try:
            self.retrieve(pf_project_id)
        except HTTPError:
            return False
        return True

    def delete_project(self, pf_project_id: UUID) -> None:
        """Delete a project."""
        self.delete(pk=pf_project_id)

    def list_users(self, pf_project_id: UUID) -> List[Dict[str, Any]]:
        """List all project members."""
        users = self.list(
            path=f"{pf_project_id}/pf_user",
            pagination=True,
        )
        return users


class ProjectCredentialClient(Client, ProjectRequestMixin):
    """Project credential client."""

    def __init__(self, **kwargs):
        """Initialize project credential client."""
        self.initialize_project()
        super().__init__(project_id=self.project_id, **kwargs)

    @property
    def url_path(self) -> Template:
        """Get an URL path."""
        return Template(
            self.url_provider.get_auth_uri("pf_project/$project_id/credential")
        )

    def list_credentials(
        self, cred_type: Optional[CredType] = None
    ) -> List[Dict[str, Any]]:
        """List all credentials."""
        params = {}
        if cred_type is not None:
            params["type"] = cred_type_map[cred_type]
        data = self.list(pagination=False, params=params)
        return data

    def create_credential(
        self, cred_type: CredType, name: str, type_version: int, value: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Create a new credential."""
        type_name = cred_type_map[cred_type]
        request_data = {
            "type": type_name,
            "name": name,
            "type_version": type_version,
            "value": value,
        }
        data = self.post(json=request_data)
        return data



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow GPT NeoX Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import GPTNeoXConfig  # type: ignore[import]

from periflow.errors import CheckpointConversionError, NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import (
    DECODER_PREFIX,
    SUPPORTED_GELU_FAMILY,
    DecoderOnlyConverter,
)
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    convert_to_gpt_j_params,
    get_tensor_from_state_dict,
    nontype_partial,
)


class GPTNeoXForCausalLMConverter(DecoderOnlyConverter):
    """GPTNeoXForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if GPTNeoX architectures' config can be converted to Periflow format."""
        super().check_config()
        try:
            if cast(GPTNeoXConfig, self.config).hidden_act not in SUPPORTED_GELU_FAMILY:
                raise NotSupportedCheckpointError(
                    invalid_option=f"'hidden_act={cast(GPTNeoXConfig, self.config).hidden_act}'",
                    valid_options=SUPPORTED_GELU_FAMILY,
                )
            if not cast(GPTNeoXConfig, self.config).use_parallel_residual:
                raise NotSupportedCheckpointError(
                    invalid_option="'use_parallel_residual=False'",
                    valid_options=[True],
                )
            if cast(GPTNeoXConfig, self.config).tie_word_embeddings:
                raise NotSupportedCheckpointError(
                    invalid_option="'tie_word_embeddings=True'",
                    valid_options=[False],
                )
            if cast(GPTNeoXConfig, self.config).layer_norm_eps != 1e-5:
                raise NotSupportedCheckpointError(
                    invalid_option="'layer_norm_eps="
                    f"{cast(GPTNeoXConfig, self.config).layer_norm_eps}'",
                    valid_options=[1e-5],
                )
            if cast(GPTNeoXConfig, self.config).rotary_emb_base != 10000:
                raise NotSupportedCheckpointError(
                    invalid_option=(
                        f"'rotary_emb_base={cast(GPTNeoXConfig, self.config).rotary_emb_base}'"
                    ),
                    valid_options=[10000],
                )
        except AttributeError as exc:
            raise CheckpointConversionError(str(exc)) from exc

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: str,  # type: ignore[override]
    ) -> np.ndarray:
        """qkv_weight_convert for GPTNeoX's attention layer."""
        assert len(per_layer_postfixes) == 1
        qkv_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        qkv_weight = qkv_weight.reshape(
            self.decoder_num_attention_heads,
            3,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )

        q_weight = qkv_weight[:, 0].reshape(
            self.decoder_num_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = qkv_weight[:, 1].reshape(
            self.decoder_num_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        v_weight = qkv_weight[:, 2].reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        q_weight = convert_to_gpt_j_params(param=q_weight, rotary_dim=self.rotary_dim)
        k_weight = convert_to_gpt_j_params(param=k_weight, rotary_dim=self.rotary_dim)
        q_weight = q_weight.reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        qkv_weight = torch.cat((q_weight, k_weight, v_weight), dim=0)
        qkv_weight = qkv_weight.transpose(0, 1)

        return convert_tensor_to_np_array(qkv_weight, self.data_type)

    def qkv_bias_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: str,  # type: ignore[override]
    ) -> np.ndarray:
        """qkv_bias_convert for GPTNeoX's attention layer."""
        assert len(per_layer_postfixes) == 1
        qkv_bias = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )
        qkv_bias = qkv_bias.reshape(
            self.decoder_num_attention_heads,
            3,
            self.decoder_head_size,
        )

        q_bias = qkv_bias[:, 0].reshape(
            self.decoder_num_attention_heads, self.decoder_head_size
        )
        k_bias = qkv_bias[:, 1].reshape(
            self.decoder_num_attention_heads, self.decoder_head_size
        )
        v_bias = qkv_bias[:, 2].reshape(
            self.decoder_num_attention_heads * self.decoder_head_size
        )

        q_bias = convert_to_gpt_j_params(q_bias, self.rotary_dim).flatten()
        k_bias = convert_to_gpt_j_params(k_bias, self.rotary_dim).flatten()

        qkv_bias = torch.cat((q_bias, k_bias, v_bias), dim=0)
        return convert_tensor_to_np_array(qkv_bias, self.data_type)

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(GPTNeoXConfig, self.config)

        logger.info(
            "The generated attributes set 'max_length' to %d, but you can change the "
            "'max_length' according to your needs. The GPTNeoX model does not rely on "
            "absolute position embeddings, allowing you to choose any suitable value.",
            config.max_position_embeddings,
        )

        eos_token_id = self.get_eos_token_id()
        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": config.max_position_embeddings,
            "vocab_size": config.vocab_size,
            "eos_token": eos_token_id if eos_token_id is not None else "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        return "gpt-neox"

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in GPTNeoX."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["gpt_neox.embed_in.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["gpt_neox.final_layer_norm.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=["gpt_neox.final_layer_norm.bias"],
            ),
            "head_fc/weight:0": nontype_partial(
                self.head_weight_convert, per_layer_postfixes=["embed_out.weight"]
            ),
        }

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in GPTNeoX."""
        return {
            "ln_1/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".input_layernorm.weight"],
            ),
            "ln_1/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".input_layernorm.bias"],
            ),
            "attn/c_attn/bias:0": nontype_partial(
                self.qkv_bias_convert,
                per_layer_postfixes=[".attention.query_key_value.bias"],
            ),
            "attn/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".attention.dense.bias"],
            ),
            "ln_2/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=[".post_attention_layernorm.weight"],
            ),
            "ln_2/beta:0": nontype_partial(
                self.ln_bias_convert,
                per_layer_postfixes=[".post_attention_layernorm.bias"],
            ),
            "mlp/c_fc/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.dense_h_to_4h.bias"],
            ),
            "mlp/c_proj/bias:0": nontype_partial(
                self.linear_bias_convert,
                per_layer_postfixes=[".mlp.dense_4h_to_h.bias"],
            ),
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[".attention.query_key_value.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".attention.dense.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_h_to_4h.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_4h_to_h.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before GPTNeoX's transformer block number."""
        return "gpt_neox.layers."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in GPTNeoX."""
        return cast(GPTNeoXConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in GPTNeoX."""
        return cast(GPTNeoXConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in GPTNeoX."""
        return cast(GPTNeoXConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in gpt_neox."""
        return self.decoder_num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head siez of GPTNeoX."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimesion of GPTNeoX."""
        return int(self.decoder_head_size * cast(GPTNeoXConfig, self.config).rotary_pct)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow OPTForCausalLM QuantizerHook."""

# mypy: ignore-errors

from __future__ import annotations

from typing import Iterator, List, Tuple, Type, cast

import torch
from transformers.models.opt import OPTForCausalLM  # type: ignore[import]

from periflow.modules.quantizer.schema.config import SmoothQuantConfig
from periflow.modules.quantizer.schema.data import ModuleName, QuantInput, TFQuantInputs
from periflow.modules.quantizer.smoothquant.base import SmoothQuantHook


class SmoothQuantOPTHook(SmoothQuantHook):
    """SmoothQuant Hook for OPTForCausalLM."""

    def iter_smooth_norm_weights(
        self, model: OPTForCausalLM
    ) -> Iterator[Tuple[List[torch.Tensor], List[torch.Tensor], ModuleName]]:
        """Returns iterator of layernorm's weight and linear layer's weight per transformer block in OPTForCausalLM."""
        quant_args = cast(SmoothQuantConfig, self.quant_config).smoothquant_args
        for index, decoder_layer in enumerate(model.model.decoder.layers):  # type: ignore[union-attr]
            # [LayerNorm 1] - [ QKV projection ] gets smoothed
            yield (
                [
                    decoder_layer.self_attn_layer_norm.weight.data,
                    decoder_layer.self_attn_layer_norm.bias.data,
                ],
                [
                    decoder_layer.self_attn.q_proj.weight.data,
                    decoder_layer.self_attn.k_proj.weight.data,
                    decoder_layer.self_attn.v_proj.weight.data,
                ],
                f"{self.quantized_layer_prefix}{index}.self_attn.q_proj",  # the input tensors fed into Q, K, V matrices are identical.
            )
            # [LayerNorm 2] - [ MLP FF 1 ] gets smoothed
            yield (
                [
                    decoder_layer.final_layer_norm.weight.data,
                    decoder_layer.final_layer_norm.bias.data,
                ],
                [decoder_layer.fc1.weight.data],
                f"{self.quantized_layer_prefix}{index}.fc1",
            )
            if quant_args.attn_fc_smoothing:
                yield (
                    [decoder_layer.attn_fc_pre_smoother.scale.data],
                    [decoder_layer.self_attn.out_proj.weight.data],
                    f"{self.quantized_layer_prefix}{index}.self_attn.out_proj",
                )
            if quant_args.ff2_smoothing:
                yield (
                    [decoder_layer.ff2_pre_smoother.scale.data],
                    [decoder_layer.fc2.weight.data],
                    f"{self.quantized_layer_prefix}{index}.fc2",
                )

    def iter_quant_inputs(self, model: OPTForCausalLM) -> Iterator[TFQuantInputs]:
        """Returns the layers which should be quantized in transformer block of OPTForCausalLM."""
        for index, decoder_layer in enumerate(model.model.decoder.layers):
            self_attn = decoder_layer.self_attn
            fc1 = decoder_layer.fc1
            fc2 = decoder_layer.fc2
            yield TFQuantInputs(
                layer_index=index,
                q=QuantInput(
                    self_attn.q_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.q_proj",
                    None,
                    None,
                ),
                k=QuantInput(
                    self_attn.k_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.k_proj",
                    None,
                    None,
                ),
                v=QuantInput(
                    self_attn.v_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.v_proj",
                    None,
                    None,
                ),
                attn_fc=QuantInput(
                    self_attn.out_proj.weight,
                    f"{self.quantized_layer_prefix}{index}.self_attn.out_proj",
                    None,
                    None,
                ),
                ff1=QuantInput(
                    fc1.weight,
                    f"{self.quantized_layer_prefix}{index}.fc1",
                    None,
                    None,
                ),
                ff2=QuantInput(
                    fc2.weight,
                    f"{self.quantized_layer_prefix}{index}.fc2",
                    None,
                    None,
                ),
            )

    def get_linear_layer_types(self) -> Tuple[Type[torch.nn.Module]]:
        """Returns the linear layer types in OPTForCausalLM."""
        return (torch.nn.Linear,)

    def get_attn_fc_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after attention in the decoder layer."""
        return decoder_layer.self_attn.out_proj

    def get_ff2_layer(self, decoder_layer: torch.nn.Module) -> torch.nn.Linear:
        """Returns the linear layer after FF1 in the decoder layer."""
        return decoder_layer.fc2

    def get_tf_blocks(self, model: OPTForCausalLM) -> List[torch.nn.Module]:
        """Returns the decoder layers(transformer blocks) in the model."""
        return list(model.model.decoder.layers)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow Falcon Checkpoint Converter."""

from __future__ import annotations

from typing import Any, Callable, Dict, List, cast

import numpy as np
import torch
from transformers import FalconConfig  # type: ignore[import]

from periflow.errors import NotSupportedCheckpointError
from periflow.logging import logger
from periflow.modules.converter.base import DECODER_PREFIX, DecoderOnlyConverter
from periflow.modules.converter.utils import (
    convert_tensor_to_np_array,
    convert_to_gpt_j_params,
    get_tensor_from_state_dict,
    nontype_partial,
)


class FalconForCausalLMConverter(DecoderOnlyConverter):
    """FalconForCausalLM Architectures Converter Class."""

    def check_config(self) -> None:
        """Check if Falcon architectures' config can be converted to PeriFlow format."""
        super().check_config()
        config = cast(FalconConfig, self.config)

        if config.layer_norm_epsilon != 1e-5:
            raise NotSupportedCheckpointError(
                invalid_option=f"'layer_norm_epsilon={config.layer_norm_epsilon}'",
                valid_options=[1e-5],
            )

        if config.alibi:
            raise NotSupportedCheckpointError(
                invalid_option=f"'alibi'={config.alibi}'",
                valid_options=[False],
            )

        if not config.rotary:
            raise NotSupportedCheckpointError(
                invalid_option=f"'rotary'={config.rotary}'",
                valid_options=[True],
            )

        if config.bias:
            raise NotSupportedCheckpointError(
                invalid_option=f"'bias'={config.bias}'",
                valid_options=[False],
            )

        if not config.new_decoder_architecture and not config.parallel_attn:
            raise NotSupportedCheckpointError(
                invalid_option=(
                    f"'new_decoder_architecture'={config.new_decoder_architecture}"
                    f"'parallel_attn'={config.parallel_attn}"
                ),
                valid_options=[
                    "'new_decoder_architecture'=True",
                    "'new_decoder_architecture'=False, 'parallel_attn'=True",
                ],
            )

    def qkv_weight_convert(
        self,
        state_dict: Dict[str, torch.Tensor],
        layer: str,
        per_layer_postfixes: List[str],
    ) -> np.ndarray:
        """qkv_weight_convert for Falcon's attention layer."""
        assert len(per_layer_postfixes) == 1
        qkv_weight = get_tensor_from_state_dict(
            state_dict, layer + per_layer_postfixes[0]
        )

        num_queries_per_kv = (
            self.decoder_num_attention_heads // self.decoder_num_kv_attention_heads
        )

        qkv_weight = qkv_weight.reshape(
            self.decoder_num_kv_attention_heads,
            num_queries_per_kv + 2,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )

        q_weight = qkv_weight[:, :num_queries_per_kv].reshape(
            self.decoder_num_kv_attention_heads * num_queries_per_kv,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = qkv_weight[:, [-2]].reshape(
            self.decoder_num_kv_attention_heads,
            self.decoder_head_size,
            self.decoder_hidden_size,
        )
        v_weight = qkv_weight[:, [-1]].reshape(
            self.decoder_num_kv_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        q_weight = convert_to_gpt_j_params(q_weight, self.rotary_dim)
        k_weight = convert_to_gpt_j_params(k_weight, self.rotary_dim)

        q_weight = q_weight.reshape(
            self.decoder_num_kv_attention_heads
            * num_queries_per_kv
            * self.decoder_head_size,
            self.decoder_hidden_size,
        )
        k_weight = k_weight.reshape(
            self.decoder_num_kv_attention_heads * self.decoder_head_size,
            self.decoder_hidden_size,
        )

        qkv_weight = torch.cat((q_weight, k_weight, v_weight), dim=0)
        qkv_weight = qkv_weight.transpose(0, 1)

        return convert_tensor_to_np_array(qkv_weight, self.data_type)

    def get_attributes(self) -> Dict[str, Any]:
        """Get checkpoint attributes."""
        config = cast(FalconConfig, self.config)

        logger.warn(
            "The 'max_length' field is left blank as it cannot be automatically configured. "
            "You must determine the 'max_length' according to your needs. The Falcon model does "
            "not rely on absolute position embeddings, allowing you to choose any "
            "suitable value."
        )

        attr = {
            "model_type": self.model_type,
            "dtype": self.data_type.value,
            "head_size": self.decoder_head_size,
            "rotary_dim": self.rotary_dim,
            "num_heads": self.decoder_num_attention_heads,
            "num_kv_heads": self.decoder_num_kv_attention_heads,
            "num_layers": self.decoder_layer_num,
            "max_length": "FILL ME",
            "vocab_size": config.vocab_size,
            "eos_token": self.get_eos_token_id() or "FILL ME",
        }
        return attr

    @property
    def model_type(self) -> str:
        """Model type."""
        if cast(FalconConfig, self.config).new_decoder_architecture:
            return "falcon"
        return "falcon-7b"

    @property
    def decoder_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for transformer blocks in Falcon."""
        convert_dict = {
            "attn/c_attn/weight:0": nontype_partial(
                self.qkv_weight_convert,
                per_layer_postfixes=[".self_attention.query_key_value.weight"],
            ),
            "attn/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".self_attention.dense.weight"],
            ),
            "mlp/c_fc/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_h_to_4h.weight"],
            ),
            "mlp/c_proj/weight:0": nontype_partial(
                self.linear_weight_convert,
                per_layer_postfixes=[".mlp.dense_4h_to_h.weight"],
            ),
        }

        if cast(FalconConfig, self.config).new_decoder_architecture:
            convert_dict.update(
                {
                    "ln_1/gamma:0": nontype_partial(
                        self.ln_weight_convert,
                        per_layer_postfixes=[".ln_attn.weight"],
                    ),
                    "ln_1/beta:0": nontype_partial(
                        self.ln_bias_convert,
                        per_layer_postfixes=[".ln_attn.bias"],
                    ),
                    "ln_2/gamma:0": nontype_partial(
                        self.ln_weight_convert,
                        per_layer_postfixes=[".ln_mlp.weight"],
                    ),
                    "ln_2/beta:0": nontype_partial(
                        self.ln_bias_convert,
                        per_layer_postfixes=[".ln_mlp.bias"],
                    ),
                }
            )
        else:
            convert_dict.update(
                {
                    "ln_1/gamma:0": nontype_partial(
                        self.ln_weight_convert,
                        per_layer_postfixes=[".input_layernorm.weight"],
                    ),
                    "ln_1/beta:0": nontype_partial(
                        self.ln_bias_convert,
                        per_layer_postfixes=[".input_layernorm.bias"],
                    ),
                }
            )

        return convert_dict

    @property
    def non_transformer_convert_dict(
        self,
    ) -> Dict[str, Callable[[Dict[str, torch.Tensor], str], np.ndarray]]:
        """The convert_dict for non-transformer blocks in Falcon."""
        return {
            "wte/weight:0": nontype_partial(
                self.token_embed_weight_convert,
                per_layer_postfixes=["transformer.word_embeddings.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/gamma:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["transformer.ln_f.weight"],
            ),
            DECODER_PREFIX
            + "/ln_f/beta:0": nontype_partial(
                self.ln_weight_convert,
                per_layer_postfixes=["transformer.ln_f.bias"],
            ),
            "head_fc/weight:0": nontype_partial(
                self.head_weight_convert,
                per_layer_postfixes=["lm_head.weight"],
            ),
        }

    @property
    def decoder_layer_prefix(self) -> str:
        """The layer name prefix used before the Falcon's transformer block number."""
        return "transformer.h."

    @property
    def decoder_layer_num(self) -> int:
        """The number of decoder layers in Falcon."""
        return cast(FalconConfig, self.config).num_hidden_layers

    @property
    def decoder_hidden_size(self) -> int:
        """The hidden size in Falcon."""
        return cast(FalconConfig, self.config).hidden_size

    @property
    def decoder_num_attention_heads(self) -> int:
        """The number of attention heads in Falcon."""
        return cast(FalconConfig, self.config).num_attention_heads

    @property
    def decoder_num_kv_attention_heads(self) -> int:
        """The number of key-value attention heads in Falcon."""
        config = cast(FalconConfig, self.config)

        if config.new_decoder_architecture:
            if config.num_kv_heads is not None:
                return config.num_kv_heads
            return config.num_attention_heads

        if config.multi_query:
            return 1

        if config.num_kv_heads is not None:
            return config.num_kv_heads
        return config.num_attention_heads

    @property
    def decoder_head_size(self) -> int:
        """The head size of Falcon."""
        return self.decoder_hidden_size // self.decoder_num_attention_heads

    @property
    def rotary_dim(self) -> int:
        """The rotary embedding dimesion of Falcon."""
        return self.decoder_head_size



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Interface for resource management SDK."""

# pylint: disable=redefined-builtin

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

R = TypeVar("R")
RI = TypeVar("RI")


class ResourceAPI(ABC, Generic[R, RI]):
    """Abstract class for resource APIs."""

    @staticmethod
    @abstractmethod
    def create(*args, **kwargs) -> R:
        """Creates a resource."""

    @staticmethod
    @abstractmethod
    def get(id: RI, *args, **kwargs) -> R:
        """Gets a specific resource."""

    @staticmethod
    @abstractmethod
    def list(*args, **kwargs) -> List[R]:
        """Lists reousrces."""



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

from __future__ import annotations

from typing import ClassVar as _ClassVar
from typing import Optional as _Optional

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message

DESCRIPTOR: _descriptor.FileDescriptor

class V1TokenizeRequest(_message.Message):
    __slots__ = ["prompt"]
    PROMPT_FIELD_NUMBER: _ClassVar[int]
    prompt: str
    def __init__(self, prompt: _Optional[str] = ...) -> None: ...



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: detokenize.proto
"""Generated protocol buffer code."""
from __future__ import annotations

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10\x64\x65tokenize.proto"%\n\x13V1DetokenizeRequest\x12\x0e\n\x06tokens\x18\x02 \x03(\x05\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "detokenize_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_V1DETOKENIZEREQUEST"]._serialized_start = 20
    _globals["_V1DETOKENIZEREQUEST"]._serialized_end = 57
# @@protoc_insertion_point(module_scope)
